from typing import List, Union, Optional, Callable
import numpy as np
from tqdm import trange
import torch
import logging

from thoi.commons import _normalize_input_data
from thoi.heuristics.scoring import _evaluate_nplet_hot_encoded

def _random_solutions(repeat, N, device):
    # Create a tensor of random 0s and 1s
    current_solution = torch.randint(0, 2, (repeat, N), device=device)

    # Ensure at least 3 ones in each element
    # Randomly choose 3 unique positions for each batch element
    rand_indices = torch.rand(repeat, N, device=device).argsort(dim=1)[:, :3]

    # Set these positions to 1
    current_solution[torch.arange(repeat).unsqueeze(1), rand_indices] = 1

    return current_solution

@torch.no_grad()
def hot_encode_to_indexes(nplets):
    batch_size, N = nplets.shape
    non_zero_indices = nplets.nonzero(as_tuple=False)
    indices = non_zero_indices[:, 1].view(batch_size, -1)
    return indices

@torch.no_grad()
def simulated_annealing_multi_order(X: Union[np.ndarray, torch.Tensor, List[np.ndarray], List[torch.Tensor]],
                                    *,
                                    covmat_precomputed: bool=False,
                                    T: Optional[Union[int, List[int]]]=None,
                                    initial_solution: Optional[torch.Tensor] = None,
                                    repeat: int = 10,
                                    batch_size: int = 1000000,
                                    device: torch.device = torch.device('cpu'),
                                    max_iterations: int = 1000,
                                    early_stop: int = 100,
                                    initial_temp: float = 100.0,
                                    cooling_rate: float = 0.99,
                                    step_size: int = 1,
                                    metric: Union[str,Callable]='o', # tc, dtc, o, s
                                    largest: bool = False,
                                    verbose: int = logging.INFO):
    
    logging.basicConfig(
        level=verbose,
        format='%(asctime)s - %(levelname)s - %(message)s'
    )

    covmats, D, N, T = _normalize_input_data(X, covmat_precomputed, T, device)
    
    # Compute current solution
    # |batch_size| x |order|
    if initial_solution is None:
        current_solution = _random_solutions(repeat, N, device)
    else:
        current_solution = initial_solution.to(device).contiguous()

    # |batch_size|
    current_energy = _evaluate_nplet_hot_encoded(covmats, T,
                                                 current_solution,
                                                 metric,
                                                 batch_size=batch_size,
                                                 device=device)

    if not largest:
        current_energy = -current_energy

    # Set initial temperature
    temp = initial_temp

    # Best solution found
    # |batch_size| x |order|
    best_solution = current_solution.clone()
    # |batch_size|
    best_energy = current_energy.clone()
    
    # Repeat tensor for indexing the current_solution
    # |repeat| x |step_size|
    i_repeat = torch.arange(repeat).unsqueeze(1).expand(-1, step_size)

    no_progress_count = 0
    pbar = trange(max_iterations, leave=True)
    for _ in pbar:

        # Get function name if metric is a function
        metric_name = metric.__name__ if callable(metric) else metric

        pbar.set_description(f'mean({metric_name}) = {(1 if largest else -1) * best_energy.mean()} - ES: {no_progress_count}')
        
        
        # |batch_size|
        #i_change = torch.randint(0, N, (repeat,), device=device)
        i_change = torch.stack([torch.randperm(N, device=device)[:step_size] for _ in range(repeat)])
        

        # Change values of selected elements
        current_solution[i_repeat, i_change] = 1 - current_solution[i_repeat, i_change]

        # Calculate energy of new solution
        # |batch_size|
        new_energy = _evaluate_nplet_hot_encoded(covmats, T,
                                                 current_solution,
                                                 metric,
                                                 batch_size=batch_size,
                                                 device=device)

        if not largest:
            new_energy = -new_energy

        # Calculate change in energy
        # delca_energy > 0 means new_energy is bigger (more optimal) than current_energy
        # |batch_size|
        delta_energy = new_energy - current_energy

        # Determine if we should accept the new solution
        # |batch_size|
        temp_probas = torch.rand(repeat, device=device) < torch.exp(delta_energy / temp)
        improves = delta_energy > 0
        accept_new_solution = torch.logical_or(improves, temp_probas)

        # valid solutions have at least three elements with 1. Non valid solution are not accepted
        valid_solutions = current_solution.sum(dim=1) >= 3
        accept_new_solution = torch.logical_and(accept_new_solution, valid_solutions)

        # revert changes of not accepted solutions in the rows of the mask accept_new_solution and in the indexes of i_change
        current_solution[i_repeat[~accept_new_solution], i_change[~accept_new_solution]] = 1 - current_solution[i_repeat[~accept_new_solution], i_change[~accept_new_solution]]

        # |batch_size|
        current_energy[accept_new_solution] = new_energy[accept_new_solution]

        # |batch_size|
        new_global_maxima = (new_energy > best_energy)

        # |batch_size| x |order|
        best_solution[new_global_maxima] = current_solution[new_global_maxima] 

        # |batch_size|
        best_energy[new_global_maxima] = new_energy[new_global_maxima]

        # Cool down
        temp *= cooling_rate

        # Early stop
        if torch.any(new_global_maxima):
            no_progress_count = 0
        else:
            no_progress_count += 1
        
        if no_progress_count >= early_stop:
            logging.info('Early stop reached')
            break

    # If minimizing, then return score to its real value
    if not largest:
        best_energy = -best_energy

    return best_solution, best_energy


from typing import Optional, Callable, Union, List

from tqdm import tqdm
from functools import partial
import logging

import numpy as np
import torch
from torch.utils.data import DataLoader
import torch.nn.functional as F

from thoi.typing import TensorLikeArray
from thoi.commons import _normalize_input_data
from thoi.dataset import CovarianceDataset
from thoi.collectors import batch_to_csv, concat_and_sort_csv
from thoi.measures.gaussian_copula_hot_encoded import nplets_measures_hot_encoded
from thoi.measures.utils import _all_min_1_ids, \
                                _multivariate_gaussian_entropy, \
                                _univariate_gaussian_entropy, \
                                _marginal_gaussian_entropies, \
                                _gaussian_entropy_bias_correction, \
                                _get_single_exclusion_covmats

def _indices_to_hot_encoded(nplets_idxs, N):
    """
    Converts a list of index lists to a multi-hot encoded tensor using one_hot and masking.

    Parameters:
    - nplets_idxs: List of lists of indices.
    - N: Number of classes.

    Returns:
    - multi_hot: torch.Tensor of shape (batch_size, N)
    """
    return torch.stack([
        F.one_hot(torch.as_tensor(lst), num_classes=N).sum(dim=0)
        for lst in nplets_idxs
    ])


def _generate_nplets_marginal_entropies(marginal_entropies: torch.Tensor, nplets: torch.Tensor):
    
    D, N = marginal_entropies.shape
    batch_size, order = nplets.shape
    
    # Expand entropies
    # |batch_size| x |D| x |N|
    entropies_expanded = marginal_entropies.unsqueeze(0).expand(batch_size, D, N)

    # Expand nplets and repeat them across the D dimensions
    # |batch_size| x |D| x |order|
    nplets_expanded = nplets.unsqueeze(1).expand(batch_size, D, order)

    # Gather the entropies based on nplets indices
    # |batch_size| x |D| x |order|
    nplets_marginal_entropies = torch.gather(entropies_expanded, dim=2, index=nplets_expanded)

     # |batch_size| x |D| x |order|
    return nplets_marginal_entropies
    

def _generate_nplets_covmants(covmats: torch.Tensor, nplets: torch.Tensor):
    
    batch_size, order = nplets.shape
    D, N = covmats.shape[:2]
    
    # Expand nplets to match the dimensions needed for batch indexing
    # |batch_size| x |D| x |order|
    nplets_expanded = nplets.unsqueeze(1).expand(-1, D, -1)

    # Prepare covmats for batch indexing to gather elements along the N dimension
    # |batch_size| x |D| x |N| x |N|
    covmats_expanded = covmats.unsqueeze(0).expand(batch_size, -1, -1, -1)

    # Gather the rows
    # |batch_size| x |D| x |order| x |N|
    indices_row = nplets_expanded.unsqueeze(-1).expand(-1, -1, -1, N)
    gathered_rows = torch.gather(covmats_expanded, 2, indices_row)

    # Gather the columns
    # |batch_size| x |D| x |order| x |order|
    indices_col = nplets_expanded.unsqueeze(-2).expand(-1, -1, order, -1)
    nplets_covmat = torch.gather(gathered_rows, 3, indices_col)

    # |batch_size| x |D| x |order| x |order|
    return nplets_covmat


def _get_bias_correctors(T: Optional[List[int]], order: int, batch_size: int, D: int, device: torch.device):
    if T is not None:
        # |batch_size|
        bc1 = torch.tensor([_gaussian_entropy_bias_correction(1,t) for t in T], device=device)
        bcN = torch.tensor([_gaussian_entropy_bias_correction(order,t) for t in T], device=device)
        bcNmin1 = torch.tensor([_gaussian_entropy_bias_correction(order-1,t) for t in T], device=device)
    else:
        # |batch_size|
        bc1 = torch.tensor([0] * D, device=device)
        bcN = torch.tensor([0] * D, device=device)
        bcNmin1 = torch.tensor([0] * D, device=device)

    # |batch_size x D|
    bc1 = bc1.repeat(batch_size)
    bcN = bcN.repeat(batch_size)
    bcNmin1 = bcNmin1.repeat(batch_size)

    return bc1, bcN, bcNmin1


def _get_tc_dtc_from_batched_covmat(covmats: torch.Tensor,
                                    allmin1: torch.Tensor,
                                    bc1: torch.Tensor,
                                    bcN: torch.Tensor,
                                    bcNmin1: torch.Tensor,
                                    marginal_entropies: Optional[torch.Tensor] = None):

    '''
    Brief: Compute the total correlation (tc), dual total correlation (dtc), o-information (o) and s-information (s) for the given batch of covariance matrices.
    
    Parameters:
    - covmats (torch.Tensor): The covariance matrices with shape (batch_size, N, N)
    - allmin1 (torch.Tensor): The indexes of marginal covariance matrices with shape (batch_size, N, N-1)
    - bc1 (torch.Tensor): The bias corrector for the first order with shape (batch_size)
    - bcN (torch.Tensor): The bias corrector for the order with shape (batch_size)
    - bcNmin1 (torch.Tensor): The bias corrector for the order-1 with shape (batch_size)
    - marginal_entropies (Optional[torch.Tensor]): The marginal entropies for each variable with shape (batch_size, N). If None, it will be dynamically computed
    '''

    N = covmats.shape[1]

    # Compute the entire system entropy
    # |batch_size|
    sys_ent = _multivariate_gaussian_entropy(covmats, N) - bcN
    
    # Compute the single variables entropy
    # |batch_size| x |N|
    if marginal_entropies is None:
        single_var_variances = torch.diagonal(covmats, dim1=-2, dim2=-1)
        marginal_entropies = _univariate_gaussian_entropy(single_var_variances)
    marginal_entropies.sub_(bc1.unsqueeze(1))
    
    # Compute the single exclusion entropies
    # |batch_size| x |N|
    single_exclusion_covmats = _get_single_exclusion_covmats(covmats, allmin1)
    single_exclusion_ents = _multivariate_gaussian_entropy(single_exclusion_covmats, N-1) - bcNmin1.unsqueeze(1)

    # |batch_size|
    nplet_tc = torch.sum(marginal_entropies, dim=1) - sys_ent
    # TODO: inf - inf return NaN in pytorch. Check how should I handle this.
    # |batch_size|
    nplet_dtc = torch.sum(single_exclusion_ents, dim=1) - (N-1.0)*sys_ent

    # |batch_size|
    nplet_o = nplet_tc - nplet_dtc
    # |batch_size|
    nplet_s = nplet_tc + nplet_dtc

    return nplet_tc, nplet_dtc, nplet_o, nplet_s

@torch.no_grad()
def nplets_measures(X: Union[TensorLikeArray],
                    nplets: Optional[TensorLikeArray] = None,
                    *,
                    covmat_precomputed: bool = False,
                    T: Optional[Union[int, List[int]]] = None,
                    device: torch.device = torch.device('cpu'),
                    verbose: int = logging.INFO,
                    batch_size: int = 1000000):
    
    """
    Compute higher-order measures (TC, DTC, O, S) for specified n-plets in the given data matrices X.

    The computed measures are:
        - **Total Correlation (TC)**
        - **Dual Total Correlation (DTC)**
        - **O-information (O)**
        - **S-information (S)**

    Parameters
    ----------
    X : TensorLikeArray
        Input data, which can be one of the following:
        - A single torch.Tensor or np.ndarray with shape (T, N).
        - A sequence (e.g., list) of torch.Tensor or np.ndarray, each with shape (T, N), representing multiple datasets.
        - A sequence of sequences, where each inner sequence is an array-like object of shape (T, N).
        If `covmat_precomputed` is True, X should be:
        - A single torch.Tensor or np.ndarray covariance matrix with shape (N, N).
        - A sequence of covariance matrices, each with shape (N, N).

    nplets : TensorLikeArray, optional
        The n-plets to calculate the measures, with shape `(n_nplets, order)`. If `None`, all possible n-plets of the given order are considered.

    covmat_precomputed : bool, optional
        If True, X is treated as covariance matrices instead of raw data. Default is False.

    T : int or list of int, optional
        Number of samples used to compute bias correction. This parameter is used only if `covmat_precomputed` is True.
        If X is a sequence of covariance matrices, T should be a list of sample sizes corresponding to each matrix.
        If T is None and `covmat_precomputed` is True, bias correction is not applied. Default is None.

    device : torch.device, optional
        Device to use for computation. Default is torch.device('cpu').

    verbose : int, optional
        Logging verbosity level. Default is `logging.INFO`.

    batch_size : int, optional
        Batch size for processing n-plets. Default is 1,000,000.

    Returns
    -------
    torch.Tensor
        Tensor containing the computed measures for each n-plet with shape `(n_nplets, D, 4)`

    Where
    -----
    D : int
        Number of datasets. If `X` is a single dataset, `D = 1`.

    N : int
        Number of variables (features) in each dataset.

    T : int
        Number of samples in each dataset (if applicable).

    order : int
        The size of the n-plets being analyzed.

    n_nplets : int
        Number of n-plets processed.

    Examples
    --------
    **Compute measures for all possible 3-plets in a single dataset:**

    ```python
    import torch
    import numpy as np

    # Sample data matrix with 100 samples and 5 variables
    X = np.random.randn(100, 5)

    # Compute measures for all 3-plets
    measures = nplets_measures(X, nplets=None, covmat_precomputed=False, T=100)
    ```

    **Compute measures for specific n-plets in multiple datasets:**

    ```python
    import torch
    import numpy as np

    # Sample data matrices for 2 datasets, each with 100 samples and 5 variables
    X1 = np.random.randn(100, 5)
    X2 = np.random.randn(100, 5)
    X = [X1, X2]

    # Define specific n-plets to analyze
    nplets = torch.tensor([[0, 1, 2], [1, 2, 3]])

    # Compute measures for the specified n-plets
    measures = nplets_measures(X, nplets=nplets, covmat_precomputed=False, T=[100, 100])
    ```

    **Compute measures with precomputed covariance matrices:**

    ```python
    import torch
    import numpy as np

    # Precompute covariance matrices for 2 datasets
    covmat1 = np.cov(np.random.randn(100, 5), rowvar=False)
    covmat2 = np.cov(np.random.randn(100, 5), rowvar=False)
    X = [covmat1, covmat2]

    # Number of samples for each covariance matrix
    T = [100, 100]

    # Define specific n-plets to analyze
    nplets = torch.tensor([[0, 1], [2, 3]])

    # Compute measures using precomputed covariance matrices
    measures = nplets_measures(X, nplets=nplets, covmat_precomputed=True, T=T)
    ```

    **Notes**
    -----
    - If `nplets` is `None`, the function considers all possible n-plets of the specified order within the datasets.
    - Ensure that the length of `T` matches the number of datasets when `covmat_precomputed` is `True` and `X` is a sequence of covariance matrices.
    - The function is optimized for batch processing using PyTorch tensors, facilitating efficient computations on large datasets.
    
    References
    ----------
    .. [1] Rosas, Fernando E., et al. "Quantifying high-order interdependencies via multivariate extensions of the mutual information." Physical Review E 100.3 (2019): 032305.

    """
    
    logging.basicConfig(
        level=verbose,
        format='%(asctime)s - %(levelname)s - %(message)s'
    )
    
    covmats, D, N, T = _normalize_input_data(X, covmat_precomputed, T, device)
    
    # If nplets is a list of nplets with different orders, then use hot encoding to compute multiorder measures
    if isinstance(nplets, list) and not all([len(nplet) == len(nplets[0]) for nplet in nplets]):
        logging.warning('Using hot encoding to compute multi-order measures as nplets have different orders')
        nplets = _indices_to_hot_encoded(nplets, N)
        return nplets_measures_hot_encoded(covmats, nplets, covmat_precomputed=True, T=T)
    elif nplets is None:
        nplets = torch.arange(N, device=device).unsqueeze(0)
    
    # If nplets are not tensors, convert to tensor
    nplets = torch.as_tensor(nplets).to(device).contiguous()
        
    # nplets must be a batched tensor
    assert len(nplets.shape) == 2, 'nplets must be a batched tensor with shape (batch_size, order)'
    batch_size = min(batch_size, len(nplets))
    order = nplets.shape[1]

    # Create marginal indexes
    # |N| x |N-1|
    allmin1 = _all_min_1_ids(order, device=device)

    # Create bias corrector values
    # |batch_size x D|, |batch_size x D|, |batch_size x D|
    bc1, bcN, bcNmin1 = _get_bias_correctors(T, order, batch_size, D, device)

    # Create DataLoader for nplets
    dataloader = DataLoader(nplets, batch_size=batch_size, shuffle=False)

    results = []
    for nplet_batch in tqdm(dataloader, desc='Processing n-plets', leave=False):
        curr_batch_size = nplet_batch.shape[0]

        # Create the covariance matrices for each nplet in the batch
        # |curr_batch_size| x |D| x |order| x |order|
        nplets_covmats = _generate_nplets_covmants(covmats, nplet_batch)
        
        # Pack covmats in a single batch
        # |curr_batch_size x D| x |order| x |order|
        nplets_covmats = nplets_covmats.view(curr_batch_size * D, order, order)

        # Batch process all nplets at once
        measures = _get_tc_dtc_from_batched_covmat(nplets_covmats,
                                                   allmin1,
                                                   bc1[:curr_batch_size * D],
                                                   bcN[:curr_batch_size * D],
                                                   bcNmin1[:curr_batch_size * D])

        # Unpack results
        # |curr_batch_size x D|, |curr_batch_size x D|, |curr_batch_size x D|, |curr_batch_size x D|
        nplets_tc, nplets_dtc, nplets_o, nplets_s = measures

        # Collect results
        results.append(torch.stack([nplets_tc.view(curr_batch_size, D),
                                    nplets_dtc.view(curr_batch_size, D),
                                    nplets_o.view(curr_batch_size, D),
                                    nplets_s.view(curr_batch_size, D)], dim=-1))

    # Concatenate all results
    return torch.cat(results, dim=0)

@torch.no_grad()
def multi_order_measures(X: TensorLikeArray,
                         min_order: int=3,
                         max_order: Optional[int]=None,
                         *,
                         covmat_precomputed: bool=False,
                         T: Optional[Union[int, List[int]]]=None,
                         batch_size: int = 1000000,
                         device: torch.device = torch.device('cpu'),
                         num_workers: int = 0,
                         batch_aggregation: Optional[Callable[[any],any]] = None,
                         batch_data_collector: Optional[Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray],any]] = None):
    """
    Compute multi-order measures (TC, DTC, O, S) for the given data matrix X.

    The measurements computed are:
        - Total Correlation (TC)
        - Dual Total Correlation (DTC)
        - O-information (O)
        - S-information (S)

    Parameters
    ----------
    X : TensorLikeArray
        Input data, which can be one of the following:
        - A single torch.Tensor or np.ndarray with shape (T, N).
        - A sequence (e.g., list) of torch.Tensor or np.ndarray, each with shape (T, N), representing multiple datasets.
        - A sequence of sequences, where each inner sequence is an array-like object of shape (T, N).
        If `covmat_precomputed` is True, X should be:
        - A single torch.Tensor or np.ndarray covariance matrix with shape (N, N).
        - A sequence of covariance matrices, each with shape (N, N).

    min_order : int, optional
        Minimum order to compute. Default is 3. Note: 3 <= min_order <= max_order <= N.
    max_order : int, optional
        Maximum order to compute. If None, uses N (number of variables). Default is None. Note: min_order <= max_order <= N.
    covmat_precomputed : bool, optional
        If True, X is treated as covariance matrices instead of raw data. Default is False.
    T : int or list of int, optional
        Number of samples used to compute bias correction. This parameter is used only if `covmat_precomputed` is True.
        If X is a sequence of covariance matrices, T should be a list of sample sizes corresponding to each matrix.
        If T is None and `covmat_precomputed` is True, bias correction is not applied. Default is None.
    batch_size : int, optional
        Batch size for DataLoader. Default is 1,000,000.
    device : torch.device, optional
        Device to use for computation. Default is torch.device('cpu').
    num_workers : int, optional
        Number of workers for DataLoader. Default is 0.
    batch_aggregation : callable, optional
        Function to aggregate the collected batch data into the final result.
        It should accept a list of outputs from `batch_data_collector` and return the final aggregated result.
        The return type of this function determines the return type of `multi_order_measures`.
        By default, it uses `concat_and_sort_csv`, which concatenates CSV data and sorts it, returning a pandas DataFrame. 
        For more information see :ref:`collectors__concat_and_sort_csv`
    batch_data_collector : callable, optional
        Function to process and collect data from each batch.
        It should accept the following parameters:
            - nplets: torch.Tensor of n-plet indices, shape (batch_size, order)
            - nplets_tc: torch.Tensor of total correlation values, shape (batch_size, D)
            - nplets_dtc: torch.Tensor of dual total correlation values, shape (batch_size, D)
            - nplets_o: torch.Tensor of O-information values, shape (batch_size, D)
            - nplets_s: torch.Tensor of S-information values, shape (batch_size, D)
            - batch_number: int, the current batch number
        The output of `batch_data_collector` must be compatible with the input expected by `batch_aggregation`.
        By default, it uses `batch_to_csv`, which collects data into CSV. For more information see :ref:`collectors__batch_to_csv`

    Returns
    -------
    Any
        The aggregated result of the computed measures. The exact type depends on the `batch_aggregation` function used.
        By default, it returns a pandas DataFrame containing the computed metrics (DTC, TC, O, S), the n-plets indexes, 
        the order and the dataset information.
    
    Where
    -----
    D : int
        Number of datasets. If X is a single dataset, D = 1.
    N : int
        Number of variables (features) in each dataset.
    T : int
        Number of samples in each dataset (if applicable).
    order : int
        The size of the n-plets being analyzed, ranging from `min_order` to `max_order`.
    batch_size : int
        Number of n-plets processed in each batch.

    Notes
    -----
    - The default `batch_data_collector` and `batch_aggregation` functions are designed to work together.
      If you provide custom functions, ensure that the output of `batch_data_collector` is compatible with the input of `batch_aggregation`.
    - Ensure that the length of `T` matches the number of datasets when `covmat_precomputed` is `True` and `X` is a sequence of covariance matrices.
    - The function computes measures for all combinations of variables of orders ranging from `min_order` to `max_order`.
    - The function is optimized for batch processing using PyTorch tensors, facilitating efficient computations on large datasets.

    Examples
    --------
    Using default batch data collector and aggregation:

    >>> result = multi_order_measures(X, min_order=3, max_order=5)

    Using custom batch data collector and aggregation:

    >>> def custom_batch_data_collector(nplets, tc, dtc, o, s, batch_number):
    ...     # Custom processing
    ...     return custom_data
    ...
    >>> def custom_batch_aggregation(batch_data_list):
    ...     # Custom aggregation
    ...     return final_result
    ...
    >>> result = multi_order_measures(
    ...     X,
    ...     min_order=3,
    ...     max_order=5,
    ...     batch_data_collector=custom_batch_data_collector,
    ...     batch_aggregation=custom_batch_aggregation
    ... )

    References
    ----------
    .. [1] Rosas, Fernando E., et al. "Quantifying high-order interdependencies via multivariate extensions of the mutual information." Physical Review E 100.3 (2019): 032305.

    """

    covmats, D, N, T = _normalize_input_data(X, covmat_precomputed, T, device)
    
    # For each dataset, precompute the single variable marginal gaussian entropies
    # |D| x |N|
    marginal_entropies = _marginal_gaussian_entropies(covmats)

    max_order = N if max_order is None else max_order

    if batch_aggregation is None:
        batch_aggregation = concat_and_sort_csv

    if batch_data_collector is None:
        batch_data_collector = partial(batch_to_csv, N=N)

    assert max_order <= N, f"max_order must be lower or equal than N. {max_order} > {N})"
    assert min_order <= max_order, f"min_order must be lower or equal than max_order. {min_order} > {max_order}"

    # Ensure that final batch_size is smaller than the original batch_size 
    batch_size = max(batch_size // D, 1)

    # To compute using pytorch, we need to compute each order separately
    batched_data = []
    for order in tqdm(range(min_order, max_order+1), leave=False, desc='Order', disable=(min_order==max_order)):

        # Calculate constant values valid for all n-plets of the current order
        # |N| x |N-1|
        allmin1 = _all_min_1_ids(order, device=device)
        
        # Create the bias corrector for the current order
        # |batch_size x D|, |batch_size x D|, |batch_size x D|
        bc1, bcN, bcNmin1 = _get_bias_correctors(T, order, batch_size, D, device)

        # Generate dataset iterable
        dataset = CovarianceDataset(N, order, device=device)
        dataloader = DataLoader(
            dataset,
            batch_size=min(batch_size, len(dataset)),
            shuffle=False,
            num_workers=num_workers
        )

        # calculate measurments for each batch
        for bn, nplets in enumerate(tqdm(dataloader, total=len(dataloader), leave=False, desc='Batch')):
            # Batch_size can vary in the last batch, then effective batch size is the current batch size
            curr_batch_size = nplets.shape[0]

            # Send nplets to the device in case it is not there
            # |curr_batch_size| x |order|
            nplets = nplets.to(device)

            # Create the covariance matrices for each nplet in the batch
            # |curr_batch_size| x |D| x |order| x |order|
            nplets_covmats = _generate_nplets_covmants(covmats, nplets)
            
            # Create the marginal entropies sampling from the marginal_entropies tensor
            # |curr_batch_size| x |D| x |order|
            nplets_marginal_entropies = _generate_nplets_marginal_entropies(marginal_entropies, nplets)

            # Pack covmats and marginal entropies in a single batch
            # |curr_batch_size x D| x |N| x |N|
            nplets_covmats = nplets_covmats.view(curr_batch_size*D, order, order)
            nplets_marginal_entropies = nplets_marginal_entropies.view(curr_batch_size*D, order)

            # Batch process all nplets at once
            measures = _get_tc_dtc_from_batched_covmat(nplets_covmats,
                                                       allmin1,
                                                       bc1[:curr_batch_size*D],
                                                       bcN[:curr_batch_size*D],
                                                       bcNmin1[:curr_batch_size*D],
                                                       nplets_marginal_entropies)

            # Unpack results
            # |curr_batch_size x D|, |curr_batch_size x D|, |curr_batch_size x D|, |curr_batch_size x D|
            nplets_tc, nplets_dtc, nplets_o, nplets_s = measures

            # Collect batch data
            data = batch_data_collector(nplets,
                                        nplets_tc.view(curr_batch_size, D),
                                        nplets_dtc.view(curr_batch_size, D),
                                        nplets_o.view(curr_batch_size, D),
                                        nplets_s.view(curr_batch_size, D),
                                        bn)

            # Append to batched data
            batched_data.append(data)

    # Aggregate all data
    return batch_aggregation(batched_data)



from typing import Optional, List, Union, Callable
import numpy as np
from tqdm import trange
import torch
import logging

from thoi.heuristics.scoring import _evaluate_nplets
from thoi.heuristics.commons import _get_valid_candidates
from thoi.commons import _normalize_input_data

@torch.no_grad()
def random_sampler(N:int, order:int, repeat:int, device:Optional[torch.device]=None):

    device = torch.device('cpu') if device is None else device

    return torch.stack([
        torch.randperm(N, device=device)[:order]
        for _ in range(repeat)
    ])

@torch.no_grad()
def simulated_annealing(X: Union[np.ndarray, torch.Tensor, List[np.ndarray], List[torch.Tensor]],
                        order: Optional[int]=None,
                        *,
                        covmat_precomputed: bool=False,
                        T: Optional[Union[int, List[int]]]=None,
                        initial_solution: Optional[torch.Tensor] = None,
                        repeat: int = 10,
                        batch_size: int = 1000000,
                        device: torch.device = torch.device('cpu'),
                        max_iterations: int = 1000,
                        early_stop: int = 100,
                        initial_temp: float = 100.0,
                        cooling_rate: float = 0.99,
                        metric: Union[str,Callable]='o', # tc, dtc, o, s
                        largest: bool = False,
                        verbose: int = logging.INFO):
    
    logging.basicConfig(
        level=verbose,
        format='%(asctime)s - %(levelname)s - %(message)s'
    )
    
    covmats, D, N, T = _normalize_input_data(X, covmat_precomputed, T, device)

    # Compute current solution
    # |repeat| x |order|
    if initial_solution is None:
        current_solution = random_sampler(N, order, repeat, device)
    else:
        current_solution = initial_solution.to(device).contiguous()

    # |repeat|
    current_energy = _evaluate_nplets(covmats, T,
                                      current_solution,
                                      metric,
                                      batch_size=batch_size,
                                      device=device)

    if not largest:
        current_energy = -current_energy

    # Initial valid candidates
    # |repeat| x |N-order|
    valid_candidates = _get_valid_candidates(current_solution, N, device)

    # Set initial temperature
    temp = initial_temp

    # Best solution found
    # |repeat| x |order|
    best_solution = current_solution.clone()
    # |repeat|
    best_energy = current_energy.clone()
    
    # Repeat tensor for indexing the current_solution
    # |repeat| x |1|
    i_repeat = torch.arange(repeat, device=device)

    no_progress_count = 0
    pbar = trange(max_iterations, leave=False)
    for _ in pbar:
        
        # Get function name if metric is a function
        metric_name = metric.__name__ if callable(metric) else metric

        pbar.set_description(f'mean({metric_name}) = {(1 if largest else -1) * best_energy.mean()} - ES: {no_progress_count}')
        
        # Generate new solution by modifying the current solution.
        # Generate the random indexes to change.
        # |repeat| x |order|
        i_sol = torch.randint(0, current_solution.shape[1], (repeat,), device=device)
        i_cand = torch.randint(0, valid_candidates.shape[1], (repeat,), device=device)

        # Update current values by new candidates and keep the original 
        # candidates to restore where the new solution is not accepted.
        current_candidates = current_solution[i_repeat, i_sol]
        new_candidates = valid_candidates[i_repeat, i_cand]
        current_solution[i_repeat, i_sol] = new_candidates

        # Calculate energy of new solution
        # |repeat|
        new_energy = _evaluate_nplets(covmats, T,
                                      current_solution,
                                      metric,
                                      batch_size=batch_size,
                                      device=device)

        if not largest:
            new_energy = -new_energy

        # Calculate change in energy
        # delca_energy > 0 means new_energy is bigger (more optimal) than current_energy
        # |repeat|
        delta_energy = new_energy - current_energy

        # Determine if we should accept the new solution
        # |repeat|
        temp_probas = torch.rand(repeat, device=device) < torch.exp(delta_energy / temp)
        improves = delta_energy > 0
        accept_new_solution = torch.logical_or(improves, temp_probas)
        
        # Restore original values for rejected candidates
        # |repeat| x |order|
        current_solution[i_repeat[~accept_new_solution], i_sol[~accept_new_solution]] = current_candidates[~accept_new_solution]
        
        # Update valid_candidate for the accepted answers as they are not longer valid candidates
        # |repeat| x |N-order|
        valid_candidates[i_repeat[accept_new_solution], i_cand[accept_new_solution]] = current_candidates[accept_new_solution]

        # Update current energy for the accepted solutions
        # |repeat|
        current_energy[accept_new_solution] = new_energy[accept_new_solution]

        new_global_maxima = (new_energy > best_energy)

        # |repeat| x |order|
        best_solution[new_global_maxima] = current_solution[new_global_maxima]

        # |repeat|
        best_energy[new_global_maxima] = new_energy[new_global_maxima]

        # Cool down
        temp *= cooling_rate

        # Early stop
        if torch.any(new_global_maxima):
            no_progress_count = 0
        else:
            no_progress_count += 1
        
        if no_progress_count >= early_stop:
            logging.info('Early stop reached')
            break

    # If minimizing, then return score to its real value
    if not largest:
        best_energy = -best_energy

    return best_solution, best_energy




import numpy as np
import scipy as sp
import torch

from thoi.measures.constants import LOGTWOPIE, TWOPIE


def _all_min_1_ids(N: int, device: torch.device=torch.device('cpu')):
    # |N| x |N|
    base_tensor = torch.arange(N, device=device).unsqueeze(0).repeat(N, 1)
    # |N| x |N|
    mask = base_tensor != torch.arange(N, device=device).unsqueeze(1)
    # |N| x |N-1|
    return base_tensor[mask].view(N, N - 1)


def _gaussian_entropy_bias_correction(N,T):
    """Computes the bias of the entropy estimator of a 
    N-dimensional multivariate gaussian with T sample"""
    psiterms = sp.special.psi((T - np.arange(1,N+1))/2)
    return torch.tensor((N*np.log(2/(T-1)) + np.sum(psiterms))/2)


def _marginal_gaussian_entropies(covmats: torch.Tensor):
    D, N = covmats.shape[:2]
    return _univariate_gaussian_entropy(
        torch.diagonal(covmats, dim1=-2, dim2=-1))


def _univariate_gaussian_entropy(variances):
    return 0.5 * (torch.log(TWOPIE * variances))


def _multivariate_gaussian_entropy(covmats, N):
    return 0.5 * (N*LOGTWOPIE + torch.logdet(covmats))


def _get_single_exclusion_covmats(covmats: torch.Tensor, allmin1: torch.Tensor):

    batch_size, N, _ = covmats.shape

    # Step 1: Expand allmin1 to match the batch size
    # |batch_size| |N| |N-1|
    allmin1_expanded = allmin1.unsqueeze(0).expand(batch_size, -1, -1)

    # Step 2: Expand covmats to include the N dimension for variable exclusion
    # |batch_size| |N| |N| |N|
    covmats_expanded = covmats.unsqueeze(1).expand(-1, N, -1, -1)

    # Step 3: Gather the rows corresponding to the indices in allmin1
    # |batch_size| |N| |N-1| |N|
    indices_row = allmin1_expanded.unsqueeze(-1).expand(-1, -1, -1, N)
    gathered_rows = torch.gather(covmats_expanded, 2, indices_row)

    # Step 4: Gather the columns corresponding to the indices in allmin1
    # Shape of indices_col: (batch_size, N, N-1, N-1)
    # |batch_size| |N| |N-1| |N-1|
    indices_col = allmin1_expanded.unsqueeze(-2).expand(-1, -1, N-1, -1)
    covmats_sub = torch.gather(gathered_rows, 3, indices_col)

    # |bz| x |N| x |N-1| x |N-1|
    return covmats_sub



from typing import List, Optional, Tuple, Union, Callable
import pandas as pd
import numpy as np
import torch
from functools import partial

from thoi.commons import _get_string_metric

####################################################################################################
######################                     To dataframe                    #########################
####################################################################################################

def batch_to_csv(nplets_idxs: torch.Tensor,
                 nplets_tc: torch.Tensor,
                 nplets_dtc: torch.Tensor,
                 nplets_o: torch.Tensor,
                 nplets_s: torch.Tensor,
                 bn:int,
                 only_synergetic: bool=False,
                 columns: Optional[List[str]]=None,
                 N: Optional[int]=None,
                 sep: str='\t',
                 indexing_method: str='indexes', # indexes or hot_encoded
                 output_path: Optional[str]=None) -> Optional[pd.DataFrame]:
    """
    Convert batch results to a pandas DataFrame and optionally save to CSV.

    This function processes the measures computed for n-plets in a batch and converts them into a pandas DataFrame.
    It can also save the DataFrame to a CSV file if an output path is provided.

    Parameters
    ----------
    nplets_idxs : torch.Tensor
        Indices of the n-plets. Shape: (batch_size, order).
    nplets_tc : torch.Tensor
        Total correlation values. Shape: (batch_size, D).
    nplets_dtc : torch.Tensor
        Dual total correlation values. Shape: (batch_size, D).
    nplets_o : torch.Tensor
        O-information values. Shape: (batch_size, D).
    nplets_s : torch.Tensor
        S-information values. Shape: (batch_size, D).
    bn : int
        Batch number, used for identification in output files.
    only_synergetic : bool, optional
        If True, only includes n-plets with negative O-information (synergetic). Default is False.
    columns : list of str, optional
        Names of the variables (features). If None, variable names will be generated as 'var_0', 'var_1', ..., 'var_N-1'.
    N : int, optional
        Total number of variables. Required if `columns` is not provided.
    sep : str, optional
        Separator to use in the CSV file. Default is tab ('\\t').
    indexing_method : str, optional
        Method used to represent n-plets. Can be 'indexes' or 'hot_encoded'. Default is 'indexes'.
    output_path : str, optional
        Path to save the CSV file. If None, the DataFrame is returned instead of being saved.

    Returns
    -------
    pd.DataFrame or None
        DataFrame containing the measures and variable information for the n-plets.
        Returns None if `output_path` is provided and the DataFrame is saved to a file.

    Where
    -----
    D : int
        Number of datasets. If measures are computed over multiple datasets, D > 1.
    N : int
        Number of variables (features).
    batch_size : int
        Number of n-plets in the batch.
    order : int
        Order of the n-plets (number of variables in each n-plet).

    Notes
    -----
    - The function can filter out n-plets with non-negative O-information if `only_synergetic` is True.
    - The resulting DataFrame includes the measures and a binary indicator for each variable indicating its presence in the n-plet.
    - The DataFrame also includes columns for 'order' and 'dataset'.

    Examples
    --------
    ```python
    # Sample inputs
    nplets_idxs = torch.tensor([[0, 1], [1, 2], [0, 2]])
    nplets_tc = torch.rand(3, 1)
    nplets_dtc = torch.rand(3, 1)
    nplets_o = torch.rand(3, 1)
    nplets_s = torch.rand(3, 1)
    bn = 0
    columns = ['A', 'B', 'C']
    N = 3

    # Convert batch to DataFrame
    df = batch_to_csv(
        nplets_idxs,
        nplets_tc,
        nplets_dtc,
        nplets_o,
        nplets_s,
        bn,
        columns=columns,
        N=N
    )
    ```
    """
    
    # nplets have shape |batch_size| x |D|
    
    assert columns is not None or N is not None, 'either columns or N must be defined'

    if N is None:
        N = len(columns)
    
    if columns is None:
        columns = [f'var_{i}' for i in range(N)]
    
    assert N == len(columns), f'N must be equal to len(columns). {N} != {len(columns)}'

    # If only_synergetic; remove nplets with nplet_o >= 0
    if only_synergetic:
        to_keep = torch.where(nplets_o < 0)[0]
        nplets_tc = nplets_tc[to_keep]
        nplets_dtc = nplets_dtc[to_keep]
        nplets_o = nplets_o[to_keep]
        nplets_s = nplets_s[to_keep]
        # TODO: Check if this is correct since to_keep has shape |batch_size| x |D| and nplets_idcx has shape |batch_size|
        nplets_idxs = nplets_idxs[to_keep.to(nplets_idxs.device)]
    
    bs, D = nplets_dtc.shape if len(nplets_dtc.shape) == 2 else (nplets_dtc.shape[0], 1)

    # One we removed not synergistic if not required on GPU, we move to CPU the
    # final results to save
    df_meas = pd.DataFrame({
        'tc': nplets_tc.detach().cpu().numpy().flatten(),
        'dtc': nplets_dtc.detach().cpu().numpy().flatten(),
        'o': nplets_o.detach().cpu().numpy().flatten(),
        's': nplets_s.detach().cpu().numpy().flatten()
    })
    
    # Create a DataFrame with the n-plets
    if indexing_method == 'indexes':
        nplets_idxs = nplets_idxs.detach().cpu().numpy()
        batch_size, order = nplets_idxs.shape
        bool_array = np.zeros((batch_size, N), dtype=bool)
        rows = np.arange(batch_size).reshape(-1, 1)
        bool_array[rows, nplets_idxs] = True
    else:
        bool_array = nplets_idxs.bool().detach().cpu().numpy()
        

    # Repeat the boolean array to match the number of datasets and store in a DataFrame
    bool_array = np.repeat(bool_array, D, axis=0)
    df_vars = pd.DataFrame(bool_array, columns=columns)

    # Concat both dataframes columns and store in disk
    df = pd.concat([df_meas, df_vars], axis=1)

    # Compute a column with the order
    df['order'] = df[columns].sum(axis=1)
    
    # Add dataset col at the first columns
    df.insert(0, 'dataset', np.tile(np.arange(D), bs))

    if output_path is not None:
        df.to_csv(output_path.format(order=order, bn=bn), index=False, sep=sep)
        return None # Don't return

    return df


def concat_and_sort_csv(batched_dataframes) -> pd.DataFrame:
    """
    Concatenate a list of DataFrames and sort them by the 'dataset' column.

    Parameters
    ----------
    batched_dataframes : list of pd.DataFrame
        List of DataFrames to concatenate and sort.

    Returns
    -------
    pd.DataFrame
        The concatenated and sorted DataFrame.

    Notes
    -----
    - The DataFrames are concatenated along the rows.
    - Sorting is performed using the 'dataset' column in ascending order.
    - The index is reset after sorting.

    Examples
    --------
    ```python
    df1 = pd.DataFrame({'dataset': [0, 0], 'value': [1, 2]})
    df2 = pd.DataFrame({'dataset': [1, 1], 'value': [3, 4]})
    combined_df = concat_and_sort_csv([df1, df2])
    ```
    """
    df = pd.concat(batched_dataframes)
    df = df.sort_values(by='dataset', kind='stable', ascending=True).reset_index(drop=True)
    return df


####################################################################################################
#######################                     To tensor                    ###########################
####################################################################################################


def top_k_nplets(nplets_idxs: torch.Tensor,
                 nplets_measures: torch.Tensor,
                 k: int,
                 metric: Union[str, Callable[[torch.Tensor], torch.Tensor]], # Add typing and documentation of the callable input
                 largest: bool) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Select the top-k n-plets based on a specified metric.

    Parameters
    ----------
    nplets_idxs : torch.Tensor
        Indices of the n-plets. Shape: (batch_size, order).
    nplets_measures : torch.Tensor
        Measures for each n-plet. Shape: (batch_size, D, 4).
    k : int
        Number of top n-plets to select.
    metric : string with value 'dtc', 'tc', 'o' or 's' or Callable
        Metric to use for ranking the n-plets. Can be a string specifying a measure ('tc', 'dtc', 'o', 's'),
        or a custom callable that takes `nplets_measures` and returns a tensor of values.
    largest : bool
        If True, selects n-plets with the largest metric values if false return n-plets with the smalest values. Default is False.

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        - Selected n-plets measures. Shape: (k, D, 4).
        - Selected n-plets indices. Shape: (k, order).
        - Metric values of the selected n-plets. Shape: (k,).

    Where
    -----
    D : int
        Number of datasets.
    order : int
        Order of the n-plets (number of variables in each n-plet).
    batch_size : int
        Number of n-plets in the batch.

    Notes
    -----
    - The function computes the specified metric for each n-plet and selects the top-k based on this metric.
    - The metric can be one of the predefined measures or a custom function.

    Examples
    --------
    ```python
    # Sample data
    nplets_idxs = torch.tensor([[0, 1], [1, 2], [0, 2]])
    nplets_measures = torch.rand(3, 1, 4)  # Assuming D=1
    k = 2
    metric = 'o'  # Use O-information for ranking

    # Get top-k n-plets
    top_measures, top_idxs, top_values = top_k_nplets(
        nplets_idxs, nplets_measures, k, metric, largest=False
    )
    ```
    """
    
    metric_func = partial(_get_string_metric, metric=metric) if isinstance(metric, str) else metric
    
    # |batch_size|
    values = metric_func(nplets_measures).to(nplets_measures.device)
    
    # |k|
    _, indices = torch.topk(values, k, largest=largest, sorted=True)

    # (|k| x |D| x |4|, |k| x |N|)
    return (
        nplets_measures[indices],
        nplets_idxs[indices.to(nplets_idxs.device)], # indices can be in CPU while the rest on GPU
        values[indices]
    )


def batch_to_tensor(nplets_idxs: torch.Tensor,
                    nplets_tc: torch.Tensor,
                    nplets_dtc: torch.Tensor,
                    nplets_o: torch.Tensor,
                    nplets_s: torch.Tensor,
                    bn: Optional[int]=None,
                    top_k: Optional[int]=None,
                    metric: Union[str, Callable[[torch.Tensor], torch.Tensor]]='o',
                    largest: bool=False)  -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
    """
    Process batch measures and optionally select top-k n-plets.

    Parameters
    ----------
    nplets_idxs : torch.Tensor
        Indices of the n-plets. Shape: (batch_size, order).
    nplets_tc : torch.Tensor
        Total correlation values. Shape: (batch_size, D).
    nplets_dtc : torch.Tensor
        Dual total correlation values. Shape: (batch_size, D).
    nplets_o : torch.Tensor
        O-information values. Shape: (batch_size, D).
    nplets_s : torch.Tensor
        S-information values. Shape: (batch_size, D).
    bn : int, optional
        Batch number. Not used in the function but kept for compatibility.
    top_k : int, optional
        If provided, selects the top-k n-plets based on the specified metric.
    metric : string with value 'dtc', 'tc', 'o' or 's' or Callable, optional
        Metric to use for ranking if `top_k` is provided. Default is 'o' (O-information).
    largest : bool, optional
        If True, selects n-plets with the largest metric values if false return n-plets with the smalest values. Default is False.

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]
        - n-plets measures. Shape: (batch_size or k, D, 4).
        - n-plets indices. Shape: (batch_size or k, order).
        - Metric values of the selected n-plets if `top_k` is provided, else None.

    Where
    -----
    D : int
        Number of datasets.
    order : int
        Order of the n-plets.
    batch_size : int
        Number of n-plets in the batch.

    Notes
    -----
    - If `top_k` is provided and less than `batch_size`, the function selects the top-k n-plets based on the metric.
    - The measures are stacked along the last dimension in the order: (tc, dtc, o, s).

    Examples
    --------
    ```python
    # Sample inputs
    nplets_idxs = torch.tensor([[0, 1], [1, 2], [0, 2]])
    nplets_tc = torch.rand(3, 1)
    nplets_dtc = torch.rand(3, 1)
    nplets_o = torch.rand(3, 1)
    nplets_s = torch.rand(3, 1)

    # Process batch without top-k selection
    measures, idxs, _ = batch_to_tensor(
        nplets_idxs, nplets_tc, nplets_dtc, nplets_o, nplets_s
    )

    # Process batch with top-2 selection based on O-information
    measures_topk, idxs_topk, metric_values = batch_to_tensor(
        nplets_idxs, nplets_tc, nplets_dtc, nplets_o, nplets_s, top_k=2, metric='o', largest=False
    )
    ```
    """

    # |batch_size| x |D|
    assert len(nplets_tc.shape) == len(nplets_dtc.shape) == len(nplets_o.shape) == len(nplets_s.shape) == 2, 'All nplets must be 2D tensors'

    # |batch_size| x |D| x |4 = (tc, dtc, o, s)|
    nplets_measures = torch.stack([nplets_tc,
                                   nplets_dtc,
                                   nplets_o,
                                   nplets_s], dim=-1)

    # If top_k is not None and len(nplets_o) > top_k, return the top_k nplets
    if (top_k is not None) and (nplets_measures.shape[0] > top_k):
        # |k x D x 4|, |k x N|, |k|
        return top_k_nplets(nplets_idxs,
                            nplets_measures,
                            top_k,
                            metric,
                            largest)

    metric_func = partial(_get_string_metric, metric=metric) if isinstance(metric, str) else metric
    
    # If not top_k or len(nplets_measuresa) > top_k  return the original values
    # |k x D x 4|, |k x N|, |k|
    values = metric_func(nplets_measures).to(nplets_measures.device)
    return (nplets_measures, nplets_idxs, values)


def concat_batched_tensors(batched_tensors: List[Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]],
                           top_k: Optional[int] = None,
                           metric: Optional[Union[str, Callable]] = 'o',
                           largest: bool = False) -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
    """
    Concatenate batched tensors and optionally select top-k n-plets.

    Parameters
    ----------
    batched_tensors : list of tuples
        Each tuple contains:
            - nplets_measures: torch.Tensor, shape (batch_size, D, 4)
            - nplets_idxs: torch.Tensor, shape (batch_size, order)
            - nplets_scores: torch.Tensor or None, shape (batch_size,)
    top_k : int, optional
        If provided, selects the top-k n-plets across all batches. Default is None.
    metric : str or Callable, optional
        Metric to use for ranking if `top_k` is provided. Default is 'o'.
    largest : bool, optional
        If True, selects n-plets with the largest metric values, if not, select n-plets with smallest values. Default is False.

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]
        - Concatenated n-plets measures. Shape: (total_nplets or k, D, 4).
        - Concatenated n-plets indices. Shape: (total_nplets or k, order).
        - Metric values of the selected n-plets if `top_k` is provided, else None.

    Where
    -----
    D : int
        Number of datasets.
    order : int
        Order of the n-plets.
    total_nplets : int
        Total number of n-plets across all batches.

    Notes
    -----
    - If `top_k` is provided, the function selects the top-k n-plets across all batches.
    - If `top_k` is provided, nplets_scores must be provided in `batched_tensors`.

    Examples
    --------
    ```python
    # Suppose we have batched tensors from two batches
    batched_tensors = [
        (measures_batch1, idxs_batch1, None),
        (measures_batch2, idxs_batch2, None)
    ]

    # Concatenate without top-k selection
    measures_all, idxs_all, _ = concat_batched_tensors(batched_tensors)

    # Concatenate and select top-5 n-plets based on O-information
    measures_topk, idxs_topk, metric_values = concat_batched_tensors(
        batched_tensors, top_k=5, metric='o', largest=False
    )
    ```
    """
    
    # Unpac the batched tensors
    nplets_measures, nplets_idxs, nplets_scores = zip(*batched_tensors)
    
    # Concatenate each tuple of tensors
    nplets_measures = torch.cat(nplets_measures)
    nplets_idxs = torch.cat(nplets_idxs)
    nplets_scores = torch.cat(nplets_scores) if nplets_scores[0] is not None else None
    
    # If top_k is not None and len(nplets_o) > top_k, return the top_k nplets
    if (top_k is not None) and (nplets_measures.shape[0] > top_k):

        assert (metric is not None) or (nplets_scores[0] is not None), 'if no metric defined to use, then nplets_scores must be defined if top_k is not None'

        if metric is not None:
            # |k x D x 4|, |k x N|, |k|
            return top_k_nplets(nplets_idxs,
                                nplets_measures,
                                top_k,
                                metric,
                                largest)

        else:

            # |k|
            _, indices = torch.topk(nplets_scores, top_k, largest=largest, sorted=True)

            # (|k| x |D| x |4|, |k| x |N|)
            return (
                nplets_measures[indices],
                nplets_idxs[indices.cpu()],
                nplets_scores[indices]
            )

    # If not top_k or len(nplets_measuresa) > top_k  return the original values
    # |k x D x 4|, |k x N|
    return (nplets_measures, nplets_idxs, nplets_scores)


from typing import Union, Callable, List, Optional
from tqdm import trange

import torch
from functools import partial

from thoi.typing import TensorLikeArray
from thoi.measures.gaussian_copula import multi_order_measures
from thoi.collectors import batch_to_tensor, concat_batched_tensors
from thoi.heuristics.commons import _get_valid_candidates
from thoi.heuristics.scoring import _evaluate_nplets
from thoi.commons import _normalize_input_data

@torch.no_grad()
def greedy(X: TensorLikeArray,
           initial_order: int=3,
           order: Optional[int]=None,
           *,
           covmat_precomputed: bool=False,
           T: Optional[Union[int, List[int]]]=None,
           repeat: int=10,
           batch_size: int=1000000,
           repeat_batch_size: int=1000000,
           device: torch.device=torch.device('cpu'),
           metric: Union[str,Callable]='o',
           largest: bool=False):

    '''
    Brief: Greedy algorithm to find the best order of nplets to maximize the metric for a given multivariate series or covariance matrices
    
    Parameters:
        - X (TensorLikeArray): The input data to compute the nplets. It can be a list of 2D numpy arrays or tensors of shape: 1. (T, N) where T is the number of samples if X are multivariate series. 2. a list of 2D covariance matrices with shape (N, N).
        - initial_order (int): The initial order to start the greedy algorithm. Default is 3.
        - order (Optional[int]): The final order to stop the greedy algorithm. If None, it will be set to N.
        - covmat_precomputed (bool): A boolean flag to indicate if the input data is a list of covariance matrices or multivariate series.
        - T (Optional[Union[int, List[int]]]): A list of integers indicating the number of samples for each multivariate series.
        - repeat (int): The number of repetition to do to obtain different solutions starting from les optimal initial solutions.
        - device (torch.device): The device to use for the computation. Default is 'cpu'.
        - batch_size (int): The batch size to use for the computation. Default is 1000000.
        - metric (Union[str,Callable]): The metric to evaluate. One of tc, dtc, o, s or a callable function
        - largest (bool): A flag to indicate if the metric is to be maximized or minimized
        
    Returns:
        - best_nplets (torch.Tensor): The nplets with the best score found with shape (repeat, order)
        - best_scores (torch.Tensor): The best scores for the best nplets with shape (repeat,)
    '''

    covmats, D, N, T = _normalize_input_data(X, covmat_precomputed, T, device)

    # Compute initial solutions
    batch_data_collector = partial(batch_to_tensor, top_k=repeat, metric=metric, largest=largest)
    batch_aggregation = partial(concat_batched_tensors, top_k=repeat, metric=None, largest=largest)

    # |repeat| x |initial_order|, |repeat|
    _, current_solution, current_scores = multi_order_measures(covmats,
                                                               covmat_precomputed=True,
                                                               T=T,
                                                               min_order=initial_order,
                                                               max_order=initial_order,
                                                               batch_size=batch_size,
                                                               device=device,
                                                               batch_data_collector=batch_data_collector,
                                                               batch_aggregation=batch_aggregation)

    # send current solution to the device
    current_solution = current_solution.to(device).contiguous()

    # Set the order to the maximum order if not specified
    order = order if order is not None else N

    # Iterate over the remaining orders to get the best solution for each order
    best_scores = [current_scores]
    for _ in trange(initial_order, order, leave=False, desc='Order'):
        
        # |repeat|, |repeat|
        best_candidate, best_score = _next_order_greedy(covmats, T, current_solution,
                                                       metric=metric,
                                                       largest=largest,
                                                       batch_size=batch_size,
                                                       repeat_batch_size=repeat_batch_size,
                                                       device=device)
        # |order - initial_order| x |repeat|
        best_scores.append(best_score)
        
        # |repeat| x |order|
        current_solution = torch.cat((current_solution, best_candidate.unsqueeze(1)) , dim=1)
    
    # |repeat| x |order|, |repeat| x |order - initial_order|
    return current_solution, torch.stack(best_scores).T


def _create_all_solutions(initial_solution: torch.Tensor, valid_candidates: torch.Tensor) -> torch.Tensor:
    """
    Concatenates initial_solution with each valid_candidate to create all_solutions.

    Parameters:
    - initial_solution (torch.Tensor): Tensor of shape (batch_size, o) containing selected indices. Note: o < N and all elements must be unique.
    - valid_candidate (torch.Tensor): Tensor of shape (batch_size, T)

    Returns:
    - all_solutions (torch.Tensor): Tensor of shape (batch_size, T, O + 1)
    """
    T = valid_candidates.shape[1]

    # Expand initial_solution to a new dimension at position 1 (for T)
    # |batch_size| x |1| x |O|
    initial_expanded = initial_solution.unsqueeze(1).expand(-1, T, -1)

    # Reshape valid_candidate to (batch_size, T, 1) for concatenation
    # |batch_size| x |T| x |1|
    valid_reshaped = valid_candidates.unsqueeze(2)

    # Concatenate along the last dimension to get (batch_size, T, O + 1)
    # |batch_size| x |T| x |O + 1|
    all_solutions = torch.cat([initial_expanded, valid_reshaped], dim=2)

    return all_solutions


def _next_order_greedy(covmats: torch.Tensor,
                      T: Optional[List[int]],
                      initial_solution: torch.Tensor,
                      metric: Union[str,Callable],
                      largest: bool,
                      batch_size: int=1000000,
                      repeat_batch_size: int=1000000,
                      device: torch.device=torch.device('cpu')):
    
    '''
    Brief: Greedy algorithm to find the best candidate to add to the current solution
    
    Parameters:
    - covmats (torch.Tensor): The covariance matrix or matrixes with shape (D, N, N)
    - T (List[int]): The number of samples for each multivariate series
    - initial_solution (torch.Tensor): The initial solution with shape (batch_size, order)
    - metric (Union[str,Callable]): The metric to evaluate. One of tc, dtc, o, s or a callable function
    - largest (bool): A flag to indicate if the metric is to be maximized or minimized
    - batch_size (int): The batch size to use for the computation. Default is 1000000.
    - device (torch.device): The device to use for the computation. Default is 'cpu'
    
    Returns:
    - best_candidates (torch.Tensor): The best candidates to add to the current solution with shape (batch_size)
    - best_score (torch.Tensor): The best score for the best candidates with shape (batch_size)
    '''

    # Get parameters attributes
    N = covmats.shape[1]
    total_size, order = initial_solution.shape

    # Initial valid candidates to iterate one by one
    # |total_size| x |N-order|
    valid_candidates = _get_valid_candidates(initial_solution, N, device)

    best_candidates = []
    best_scores = []

    for start in trange(0, total_size, repeat_batch_size, desc='Batch repeat', leave=False):
        end = min(start + repeat_batch_size, total_size)
        batch_initial_solution = initial_solution[start:end]
        batch_valid_candidates = valid_candidates[start:end]

        # |repeat_batch_size| x |N-order| x |order+1|
        all_solutions = _create_all_solutions(batch_initial_solution, batch_valid_candidates)
        
        # |repeat_batch_size x N-order| x |order+1|
        all_solutions = all_solutions.view(-1, order+1)
        
        # |repeat_batch_size x N-order|
        batch_best_score = _evaluate_nplets(covmats, T,
                                            all_solutions,
                                            metric,
                                            batch_size=batch_size,
                                            device=device)
        
        # |repeat_batch_size| x |N-order|
        batch_best_score = batch_best_score.view(end - start, N - order)
        
        if not largest:
            batch_best_score = -batch_best_score
        
        # get for each batch item the best score over the second dimension
        
        # |repeat_batch_size|
        max_idxs = torch.argmax(batch_best_score, dim=1)
        batch_best_candidates = batch_valid_candidates[torch.arange(end - start), max_idxs]
        batch_best_score = batch_best_score[torch.arange(end - start), max_idxs]
        
        # If minimizing, then return score to its original sign
        if not largest:
            batch_best_score = -batch_best_score

        best_candidates.append(batch_best_candidates)
        best_scores.append(batch_best_score)

    return torch.cat(best_candidates), torch.cat(best_scores)


from typing import Optional, List, Callable, Union

from tqdm import tqdm
from functools import partial

import numpy as np
import torch
from torch.utils.data import DataLoader

from thoi.typing import TensorLikeArray
from thoi.dataset import HotEncodedMultiOrderDataset
from thoi.collectors import batch_to_csv, concat_and_sort_csv
from thoi.measures.utils import _all_min_1_ids, \
                                _gaussian_entropy_bias_correction, \
                                _univariate_gaussian_entropy, \
                                _multivariate_gaussian_entropy, \
                                _get_single_exclusion_covmats
from thoi.measures.constants import GAUS_ENTR_NORMAL
from thoi.commons import _normalize_input_data


def _generate_nplets_covmants(covmats: torch.Tensor, nplets: torch.Tensor):

    # Ensure nplets is a float tensor
    # |batch_size| x |N|
    nplets = nplets.float()

    # Reshape covmats for broadcasting
    # |1| x |D| x |N| x |N|
    covmats_expanded = covmats.unsqueeze(0)

    # Create mask of the nplets
    # |batch_size| x |N| x |N|
    mask = nplets.unsqueeze(2) * nplets.unsqueeze(1)
    # |batch_size| x |1| x |N| x |N|
    mask = mask.unsqueeze(1)

    # Apply mask to covmats
    # |batch_size| x |D| x |N| x |N|
    masked_covmats = mask * covmats_expanded

    # Create diagonal mask for excluded variables
    # |batch_size| x |N|
    not_included = (nplets == 0).float()
    # |batch_size| x |N| x |N|
    diag_mask = torch.diag_embed(not_included)
    # |batch_size| x |1| x |N| x |N|
    diag_mask = diag_mask.unsqueeze(1)

    # Combine masked_covmats and diag_mask
    # |batch_size| x |D| x |N| x |N|
    nplets_covmat = masked_covmats + diag_mask
    
    return nplets_covmat


def _get_bias_correctors(T: Optional[List[int]], orders: torch.Tensor, batch_size: int, N: int, D: int, device: torch.device):
    
    
    # Compute bias corrector if from sampled data (T is not None)
    if T is not None:
        assert len(T) == D, 'T must have the same length as the number of datasets'
        # |N| x |D|
        bias_correctors = torch.tensor([
            [_gaussian_entropy_bias_correction(order,t) for t in T]
            for order in range(1,N+1)
        ], device=device)

    else:
        # |N| x |D|
        bias_correctors = torch.zeros((N,D), device=device)

    # |batch_size x D|
    bc1 = bias_correctors[0].repeat(batch_size)
    bcN = bias_correctors[orders-1].view(batch_size*D)
    bcNmin1 = bias_correctors[orders-2].view(batch_size*D)

    return bc1, bcN, bcNmin1


def _get_tc_dtc_from_batched_covmat(covmats: torch.Tensor, 
                                    nplets: torch.Tensor,
                                    orders: torch.Tensor,
                                    allmin1: List[np.ndarray],
                                    bc1: float, 
                                    bcOrder: float, 
                                    bcOrdermin1: float):
    '''
    Brief: Compute the total correlation (tc), dual total correlation (dtc), o-information (o) and s-information (s) for the given batch of covariance matrices.
    
    Parameters:
    - covmats (torch.Tensor): The covariance matrices with shape (batch_size*D, N, N)
    - nplets (torch.Tensor): The nplets to compute the measures with shape (batch_size, N)
    - orders (torch.Tensor): The order of each bached covmant (equivalent to covmat[...,0].sum(dim=1).int()), this must be provided to avoid multiple re calculations.
    - allmin1 (torch.Tensor): The indexes of marginal covariance matrices with shape (N, N-1)
    - bc1 (torch.Tensor): The bias corrector for the first order with shape (batch_size*D)
    - bcN (torch.Tensor): The bias corrector for the order with shape (batch_size*D)
    - bcNmin1 (torch.Tensor): The bias corrector for the order-1 with shape (batch_size*D)
    '''

    batch_size, N = covmats.shape[:2]
    D = batch_size // nplets.shape[0]
 
    # |batch_size|
    n_masked = N - orders
    
    # |batch_size| x |D| x |N|
    nplet_mask = nplets.unsqueeze(1).repeat(1, D, 1).view(batch_size, N)

    # Compute the sub covariance matrices for each variable and the system without that variable exclusion
    # |batch_size| x |N|
    single_var_covmats = torch.diagonal(covmats, dim1=-2, dim2=-1)
    # |batch_size| x |N| x |N-1| x |N-1|
    single_exclusion_covmats = _get_single_exclusion_covmats(covmats, allmin1)

    # Compute the entropy of the system, the variavbles and the system without the variable
    # |batch_size|
    sys_ent = _multivariate_gaussian_entropy(covmats, N) - bcOrder
    sys_ent.sub_(n_masked * GAUS_ENTR_NORMAL)
    # |batch_size| x |N|
    var_ents = _univariate_gaussian_entropy(single_var_covmats) - bc1.unsqueeze(1)
    var_ents.mul_(nplet_mask)

    # |batch_size| x |N|
    single_exclusion_ents = _multivariate_gaussian_entropy(single_exclusion_covmats, N-1) - bcOrdermin1.unsqueeze(1)
    single_exclusion_ents.sub_((n_masked * GAUS_ENTR_NORMAL).unsqueeze(1))
    single_exclusion_ents.mul_(nplet_mask)

    # |batch_size|
    nplet_tc = torch.sum(var_ents, dim=1) - sys_ent
    # TODO: inf - inf return NaN in pytorch. Check how should I handle this.
    # |batch_size|
    nplet_dtc = torch.sum(single_exclusion_ents, dim=1) - (orders-1.0)*sys_ent

    # |batch_size|
    nplet_o = nplet_tc - nplet_dtc
    # |batch_size|
    nplet_s = nplet_tc + nplet_dtc

    return nplet_tc, nplet_dtc, nplet_o, nplet_s


def _compute_nplets_measures_hot_encoded(covmats: torch.Tensor,
                                         T: Optional[int],
                                         N: int, D: int,
                                         nplets: Union[np.ndarray,torch.Tensor],
                                         allmin1: Optional[torch.Tensor] = None):
    batch_size = nplets.shape[0]
    device = covmats.device

    # Create the order of each nplet
    # |batch_size|
    orders = nplets.sum(dim=1).int()

    # |N| x |N-1|
    allmin1 = _all_min_1_ids(N, device=device) if allmin1 is None else allmin1

    # Create bias corrector values
    # |batch_size x D|, |batch_size x D|, |batch_size x D|
    bc1, bcN, bcNmin1 = _get_bias_correctors(T, orders, batch_size, N, D, device)

    # Create the covariance matrices for each nplet in the batch
    # |batch_size| x |D| x |N| x |N|
    nplets_covmat = _generate_nplets_covmants(covmats, nplets)

    # Pack covmats in a single batch
    # |batch_size x D| x |N| x |N|
    nplets_covmat = nplets_covmat.view(batch_size*D, N, N)

    # Batch process all nplets at once
    measures = _get_tc_dtc_from_batched_covmat(nplets_covmat,
                                               nplets,
                                               orders.repeat_interleave(D), # |batch_size x D|
                                               allmin1,
                                               bc1,
                                               bcN,
                                               bcNmin1)

    # Unpack results
    # |batch_size x D|, |batch_size x D|, |batch_size x D|, |batch_size x D|
    nplets_tc, nplets_dtc, nplets_o, nplets_s = measures
    
    return (
        nplets_tc.view(batch_size, D),
        nplets_dtc.view(batch_size, D),
        nplets_o.view(batch_size, D),
        nplets_s.view(batch_size, D)
    )

@torch.no_grad()
def nplets_measures_hot_encoded(X: TensorLikeArray,
                                nplets: Optional[TensorLikeArray] = None,
                                *,
                                covmat_precomputed: bool = False,
                                T: Optional[int] = None,
                                batch_size: int = 100000,
                                device: torch.device = torch.device('cpu')):
    '''
    Brief: Compute the higher order measurements (tc, dtc, o and s) for the given data matrices X over the nplets.
    
    Parameters:
    - X TensorLikeArray: The input data to compute the nplets. It can be a list of 2D numpy arrays or tensors of shape: 1. (T, N) where T is the number of samples if X are multivariate series. 2. a list of 2D covariance matrices with shape (N, N).
    - nplets (Optional[Union[np.ndarray,torch.Tensor]]): The nplets to calculate the measures with shape (batch_size, order)
    - covmat_precomputed (bool): A boolean flag to indicate if the input data is a list of covariance matrices or multivariate series.
    - T (Optional[Union[int, List[int]]]): A list of integers indicating the number of samples for each multivariate series.
    - device (torch.device): The device to use for the computation. Default is 'cpu'.
    - batch_size (int): Batch size for processing n-plets. Default is 100,000.
    
    Returns:
    - torch.Tensor: The measures for the nplets with shape (n_nplets, D, 4) where D is the number of matrices, n_nplets is the number of nplets to calculate over and 4 is the number of metrics (tc, dtc, o, s)
    '''

    covmats, D, N, T = _normalize_input_data(X, covmat_precomputed, T, device)

    if nplets is None:
        # If no nplets, then compute for the entire systems
        nplets = torch.ones(N, device=device).unsqueeze(0)
    else:
        # If nplets are not tensors, convert to tensor
        nplets = torch.as_tensor(nplets).to(device).contiguous()

    # nplets must be a batched tensor
    assert len(nplets.shape) == 2, 'nplets must be a batched tensor with shape (batch_size, order)'
    batch_size = min(batch_size, len(nplets))

    # Create DataLoader for nplets
    dataloader = DataLoader(nplets, batch_size=batch_size, shuffle=False)

    results = []
    for nplet_batch in tqdm(dataloader, desc='Processing n-plets', leave=False):

        # Batch process all nplets at once to obtain (tc, dtc, o, s)
        # |nplet_batch.shape[0]| x |D|, |nplet_batch.shape[0]| x |D|, |nplet_batch.shape[0] x D|, |nplet_batch.shape[0] x D|
        measures = _compute_nplets_measures_hot_encoded(covmats, T, N, D, nplet_batch)

        # Collect results
        results.append(torch.stack(measures, dim=-1))

    # Concatenate all results
    return torch.cat(results, dim=0)

@torch.no_grad()
def multi_order_measures_hot_encoded(X: TensorLikeArray,
                                     min_order: int=3,
                                     max_order: Optional[int]=None,
                                     *,
                                     covmat_precomputed: bool=False,
                                     T: Optional[int]=None,
                                     batch_size: int = 100000,
                                     device: torch.device = torch.device('cpu'),
                                     num_workers: int = 0,
                                     batch_aggregation: Optional[Callable[[any],any]] = None,
                                     batch_data_collector: Optional[Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray],any]] = None):
    """
    Compute multi-order Gaussian Copula (GC) measurements for the given data matrix X.
    The measurements computed are:
        * Total Correlation (TC)
        * Dual Total Correlation (DTC)
        * O-information (O)
        * S-information (S)

    Args:
        - X (np.ndarray or torch.Tensor): T samples x N variables matrix. if not covmat_precomputed, it should be a numpy array.
        - min_order (int): Minimum order to compute (default: 3).
        - max_order (Optional[int]): Maximum order to compute (default: None, will use N).
        - covmat_precomputed (bool): If True, X is a covariance matrix (default: False).
        - T (Optional[int]): Number of samples used to compute bias correction (default: None). This parameter is only used if covmat_precomputed is True.
        - batch_size (int): Batch size for DataLoader (default: 1000000).
        - device (torch.device): The device to use for the computation. Default is 'cpu'.
        - num_workers (int): Number of workers for DataLoader (default: 0).
        - batch_aggregation (Optional[Callable[[any],any]]): Function to aggregate the batched data (default: pd.concat).
        - batch_data_collector (Optional[Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray],any]]): Function to collect the batched data (default: batch_to_csv).

    Returns:
        pd.DataFrame: DataFrame containing computed metrics.
    """
    
    covmats, D, N, T = _normalize_input_data(X, covmat_precomputed, T, device)

    max_order = N if max_order is None else max_order

    if batch_aggregation is None:
        batch_aggregation = concat_and_sort_csv

    if batch_data_collector is None:
        batch_data_collector = partial(batch_to_csv, indexing_method='hot_encoded', N=N)

    assert max_order <= N, f"max_order must be lower or equal than N. {max_order} > {N})"
    assert min_order <= max_order, f"min_order must be lower or equal than max_order. {min_order} > {max_order}"

    batch_size = batch_size // D

    # Create marginal indexes once to be reused
    # |N| x |N-1|
    allmin1 = _all_min_1_ids(N, device=device)

    dataset = HotEncodedMultiOrderDataset(N, min_order, max_order, device=device)
    dataloader = DataLoader(
        dataset,
        batch_size=min(batch_size, len(dataset)),
        shuffle=False,
        num_workers=num_workers,
        pin_memory=device.type == 'cuda' and dataset.device.type != 'cuda'
    )

    batched_data = []
    for bn, nplets in enumerate(tqdm(dataloader, total=len(dataloader), leave=False, desc='Batch')):
        
        # nplets = |batch_size| x |N|
        
        # Batch process all nplets at once to obtain (tc, dtc, o, s)
        # |batch_size| x |D|, |batch_size| x |D|, |batch_size| x |D|, |batch_size| x |D|
        measures = _compute_nplets_measures_hot_encoded(covmats, T, N, D, nplets, allmin1)

        batched_data.append(batch_data_collector(nplets, *measures, bn))

    return batch_aggregation(batched_data)


# tests/test_multiorder_measures.py

import unittest
import numpy as np
import pandas as pd
import os

from thoi.measures.gaussian_copula import multi_order_measures
from thoi.measures.gaussian_copula_hot_encoded import multi_order_measures_hot_encoded
from thoi.commons import gaussian_copula_covmat

# TODO: make this test for all combinations of device in [cpu, cuda] and different input types
class TestMultiOrderMeasures(unittest.TestCase):

    # make the constructor
    def setUp(self):
        # Read X from the X.tsv file in the tests folder
        current_dir = os.path.dirname(__file__)
        file_path = os.path.join(current_dir, 'data','X_random.tsv')
        self.X = pd.read_csv(file_path, sep='\t', header=None).values
        self.covmat = gaussian_copula_covmat(self.X)

        # get precomputed multi order measures
        self.df_true = pd.read_csv(
            os.path.join(current_dir, 'data','X_random__multi_order_measures.tsv'),
            sep='\t'
        )

        self.cols_to_compare = ['tc', 'dtc', 'o', 's']

    def _as_sorted_dataframe(self, df):
        
        df = df.reset_index(drop=True)
        N = self.X.shape[1]
        nplet_cols = [f'var_{i}' for i in range(N)]
        df.loc[np.arange(len(df)), 'nplet'] = df[nplet_cols].apply(lambda x: ''.join(x.values.astype(int).astype(str)), axis=1)
        
        df = df.sort_values(by=['order', 'nplet'])
        df = df.reset_index(drop=True)
        return df

    def _compare_with_ground_truth(self, df_res, rtol=1e-5, atol=1e-8):
        
        df_test = self._as_sorted_dataframe(df_res)
        df_true = self._as_sorted_dataframe(self.df_true)

        pd.testing.assert_frame_equal(df_test, df_true, rtol=rtol, atol=atol)

    def _validate_same_results_for_repeated_datasets(self, df_res, rtol, atol):

        # Every dataset should have the same results as the first one
        for _, df_dataset in df_res.groupby('dataset'):
            df_dataset['dataset'] = 0 # set all datasets to 0 to compare
            self._compare_with_ground_truth(df_dataset, rtol, atol)

    def test_multiorder_measures_timeseries(self):
        df_res = multi_order_measures(self.X)
        self._compare_with_ground_truth(df_res, rtol=1e-16, atol=1e-12)
        
    def test_multiorder_measures_precomputed_covmat(self):
        df_res = multi_order_measures(self.covmat, covmat_precomputed=True, T=self.X.shape[0])
        self._compare_with_ground_truth(df_res, rtol=1e-16, atol=1e-12)

    def test_multiorder_measures_timeseries_hot_encoded(self):
        df_res = multi_order_measures_hot_encoded(self.X, batch_size=200000)
        self._compare_with_ground_truth(df_res, rtol=1e-8, atol=1e-4)

    def test_multiorder_measures_precomputed_hot_encoded(self):
        df_res = multi_order_measures_hot_encoded(self.covmat, covmat_precomputed=True, T=self.X.shape[0])
        self._compare_with_ground_truth(df_res, rtol=1e-8, atol=1e-4)

    def test_multiple_times_same_datasets_timeseries(self):
        df_res = multi_order_measures([self.X, self.X])
        self._validate_same_results_for_repeated_datasets(df_res, rtol=1e-16, atol=1e-7)
    
    def test_multiple_times_same_datasets_precomputed(self):
        covmats = [self.covmat, self.covmat]
        df_res = multi_order_measures(covmats, covmat_precomputed=True, T=self.X.shape[0])
        self._validate_same_results_for_repeated_datasets(df_res, rtol=1e-16, atol=1e-7)

    def test_multiple_times_same_datasets_timeseries_hot_encoded(self):
        df_res = multi_order_measures_hot_encoded([self.X, self.X])
        self._validate_same_results_for_repeated_datasets(df_res, rtol=1e-8, atol=1e-4)
    
    def test_multiple_times_same_datasets_precomputed_hot_encoded(self):
        covmats = [self.covmat, self.covmat]
        df_res = multi_order_measures_hot_encoded(covmats, covmat_precomputed=True, T=self.X.shape[0])
        self._validate_same_results_for_repeated_datasets(df_res, rtol=1e-8, atol=1e-4)


if __name__ == '__main__':
    unittest.main()



from torch.utils.data import IterableDataset
from itertools import combinations
import math
import numpy as np
import torch

class CovarianceDataset(IterableDataset):
    def __init__(self,
                 N: int,
                 order: int,
                 device: torch.device):

        # Depending on the resources, maybe its better to create the batches on CPU using RAM memory and/or in parallel
        self.device = device
        self.nplet_generator = combinations(range(N), order)
        self.length = math.comb(N, order)

    def __len__(self):
        """Returns the number of combinations of elements for the specified order."""
        return self.length

    def __iter__(self):
        """
        Returns the next nplet in the sequence.

        Returns:
            - nplets (torch.Tensor): The indices of the features in the current combination.
        """
        for nplet in self.nplet_generator:
            yield torch.tensor(nplet, device=self.device)


class HotEncodedMultiOrderDataset(IterableDataset):
    def __init__(self,
                 N: int,
                 min_order: int,
                 max_order: int,
                 device: torch.device):

        self.device = device
        self.N = N
        self.min_order = min_order
        self.max_order = max_order
        self.orders_length = [
            math.comb(self.N, order)
            for order in range(self.min_order, self.max_order+1)
        ]
        self.total_length = np.sum(self.orders_length)

    def __len__(self):
        return self.total_length

    def __iter__(self):
        """
        Iterate over all combinations of features in the dataset.

        Yields:
            tuple: A tuple containing:
                nplets_idxs_hot_conded (np.ndarray): A hot encoded representation of the elements in the current combination.
                partition_covmat (np.ndarray): The masked matrix where the covariance matrix corresponding to the current combination is maintained and the other rows and columns are set as independent normal (0,1) variables.
        """
        for order in range(self.min_order, self.max_order+1):
            for nplets in combinations(range(self.N), order):
                nplets_hot_encoded = torch.zeros(self.N, dtype=torch.int, device=self.device)
                nplets_hot_encoded[list(nplets)] = 1
                
                yield nplets_hot_encoded



from typing import List, Optional, Union, Callable

from thoi.commons import _get_string_metric
from thoi.measures.gaussian_copula import nplets_measures
from thoi.measures.gaussian_copula_hot_encoded import nplets_measures_hot_encoded

from functools import partial
import torch


def _evaluate_nplets(covmats: torch.Tensor,
                     T: Optional[List[int]],
                     batched_nplets: torch.Tensor,
                     metric: Union[str, Callable],
                     batch_size: int,
                     device: torch.device):
    """
        - covmats (torch.Tensor): The covariance matrix or matrixes with shape (N, N) or (D, N, N)
        - T (Optional[List[int]]): The number of samples for each multivariate series or None
        - batched_nplets (torch.Tensor): The nplets to calculate the inverse of the oinformation with shape (total_size, order)
        - metric (str): The metric to evaluate. One of tc, dtc, o, s or Callable
        - batch_size (int): The batch size to use for the computation
        - device (torch.device): The device to use
    """

    if len(covmats.shape) == 2:
        covmats = covmats.unsqueeze(0)
        
    metric_func = partial(_get_string_metric, metric=metric) if isinstance(metric, str) else metric

    # |bached_nplets| x |D| x |4 = (tc, dtc, o, s)|
    batched_measures = nplets_measures(covmats,
                                       nplets=batched_nplets,
                                       T=T,
                                       covmat_precomputed=True,
                                       batch_size=batch_size,
                                       device=device)
    
    # |batch_size|
    return metric_func(batched_measures).to(device)


def _evaluate_nplet_hot_encoded(covmats: torch.Tensor,
                                T: int,
                                batched_nplets: torch.Tensor,
                                metric: str,
                                batch_size: int,
                                device: torch.device):

    """
        covmats (torch.Tensor): The covariance matrix or matrixes with shape (N, N) or (D, N, N)
        T (Optional[List[int]]): The number of samples for each multivariate series or None
        batched_nplets (torch.Tensor): The nplets to calculate the inverse of the oinformation with shape (batch_size, order)
        metric (str): The metric to evaluate. One of tc, dtc, o, s or Callable
    """

    if len(covmats.shape) == 2:
        covmats = covmats.unsqueeze(0)

    metric_func = partial(_get_string_metric, metric=metric) if isinstance(metric, str) else metric

    # |bached_nplets| x |D| x |4 = (tc, dtc, o, s)|
    batched_measures = nplets_measures_hot_encoded(covmats,
                                                   nplets=batched_nplets,
                                                   T=T,
                                                   covmat_precomputed=True,
                                                   batch_size=batch_size,
                                                   device=device)

    # |batch_size|
    return metric_func(batched_measures).to(device)


# tests/test_multiorder_measures.py

import unittest
import numpy as np
import pandas as pd
import torch
import os

from itertools import combinations

from thoi.collectors import batch_to_csv
from thoi.measures.gaussian_copula import nplets_measures
from thoi.measures.gaussian_copula_hot_encoded import nplets_measures_hot_encoded
from thoi.commons import gaussian_copula_covmat

# TODO: make this test for all combinations of devices in [cpu, gpu] and different input types
class TestNpletsMeasures(unittest.TestCase):

    # make the constructor
    def setUp(self):
        # Read X from the X.tsv file in the tests folder
        current_dir = os.path.dirname(__file__)
        file_path = os.path.join(current_dir, 'data','X_random.tsv')
        self.X = pd.read_csv(file_path, sep='\t', header=None).values
        self.covmat = gaussian_copula_covmat(self.X)

        # get precomputed multi order measures
        self.df_true = pd.read_csv(
            os.path.join(current_dir, 'data','X_random__multi_order_measures.tsv'),
            sep='\t'
        )

        self.cols_to_compare = ['tc', 'dtc', 'o', 's']

    def _as_sorted_dataframe(self, df, nplets=None):
        
        if isinstance(df, torch.Tensor):
            assert nplets is not None
            df = batch_to_csv(nplets, df[:,:,0], df[:,:,1], df[:,:,2], df[:,:,3], 0, N=10)
        
        df = df.reset_index(drop=True)
        N = self.X.shape[1]
        nplet_cols = [f'var_{i}' for i in range(N)]
        df.loc[np.arange(len(df)), 'nplet'] = df[nplet_cols].apply(lambda x: ''.join(x.values.astype(int).astype(str)), axis=1)
        
        df = df.sort_values(by=['order', 'nplet'])
        df = df.reset_index(drop=True)
        return df

    def _compare_with_ground_truth(self, res, nplets, rtol=1e-5, atol=1e-8):
        
        df_test = self._as_sorted_dataframe(res, nplets)
        
        df_true = self.df_true[self.df_true['order'] == nplets.shape[1]]
        df_true = self._as_sorted_dataframe(df_true)

        pd.testing.assert_frame_equal(df_test, df_true, rtol=rtol, atol=atol)

    def _validate_same_results_for_repeated_datasets(self, res, nplets, rtol, atol):

        # Every dataset should have the same results as the first one
        for idx in range(res.shape[1]):
            self._compare_with_ground_truth(res[:,idx,:].unsqueeze(1), nplets, rtol, atol)

    def test_nplets_measures_timeseries(self):
        full_nplet = range(self.X.shape[1])
        for order in range(3,11):
            with self.subTest(order=order):
                nplets = torch.tensor(list(combinations(full_nplet, order)))
                res = nplets_measures(self.X, nplets)
                self._compare_with_ground_truth(res, nplets, rtol=1e-16, atol=1e-12)
    
    def test_nplets_measures_precomputed(self):
        full_nplet = range(self.X.shape[1])
        for order in range(3,11):
            with self.subTest(order=order):
                nplets = torch.tensor(list(combinations(full_nplet, order)))
                res = nplets_measures(self.covmat, nplets, covmat_precomputed=True, T=self.X.shape[0])
                self._compare_with_ground_truth(res, nplets, rtol=1e-16, atol=1e-12)
    
    def test_multiple_times_same_datasets_timeseries(self):
        full_nplet = range(self.X.shape[1])
        for order in range(3,11):
            with self.subTest(order=order):
                nplets = torch.tensor(list(combinations(full_nplet, order)))
                res = nplets_measures([self.X, self.X], nplets)
                self._validate_same_results_for_repeated_datasets(res, nplets, rtol=1e-16, atol=1e-7)
    
    def test_multiple_times_same_datasets_precomputed(self):
        full_nplet = range(self.X.shape[1])
        for order in range(3,11):
            with self.subTest(order=order):
                nplets = torch.tensor(list(combinations(full_nplet, order)))
                res = nplets_measures([self.covmat, self.covmat], nplets, covmat_precomputed=True, T=self.X.shape[0])
                self._validate_same_results_for_repeated_datasets(res, nplets, rtol=1e-16, atol=1e-7)
    
    def test_batch_size_does_not_change_result(self):
        full_nplet = range(self.X.shape[1])
        
        nplets = torch.tensor([list(c) for i, c in enumerate(combinations(full_nplet, 3)) if i < 100000])
        
        # test for different batch sizes
        res = nplets_measures(self.X, nplets)
        res2 = nplets_measures(self.X, nplets, batch_size=10)
        res3 = nplets_measures(self.X, nplets, batch_size=100)
        res4 = nplets_measures(self.X, nplets, batch_size=1000)
        res5 = nplets_measures(self.X, nplets, batch_size=10000)
        res6 = nplets_measures(self.X, nplets, batch_size=100000)
        res7 = nplets_measures(self.X, nplets, batch_size=1000000) # this should do a single batch
        
        # check that the results are the same
        self.assertTrue(torch.allclose(res, res2, rtol=1e-16, atol=1e-12))
        self.assertTrue(torch.allclose(res, res3, rtol=1e-16, atol=1e-12))
        self.assertTrue(torch.allclose(res, res4, rtol=1e-16, atol=1e-12))
        self.assertTrue(torch.allclose(res, res5, rtol=1e-16, atol=1e-12))
        self.assertTrue(torch.allclose(res, res6, rtol=1e-16, atol=1e-12))
        self.assertTrue(torch.allclose(res, res7, rtol=1e-16, atol=1e-12))

    def test_nplets_measures_timeseries_hot_encoded(self):
        N = self.X.shape[1]
        full_nplet = range(N)
        for order in range(3,11):
            with self.subTest(order=order):
                nplets = torch.tensor(list(combinations(full_nplet, order)))
                batch_size = nplets.shape[0]
                nplets_hot_encoded = torch.zeros((batch_size, N), dtype=torch.int)
                nplets_hot_encoded[torch.arange(0,batch_size, dtype=int).view(-1,1), nplets] = 1
                res = nplets_measures_hot_encoded(self.X, nplets_hot_encoded)
                self._compare_with_ground_truth(res, nplets, rtol=1e-8, atol=1e-4)

    def test_nplets_measures_precomputed_hot_encoded(self):
        N = self.X.shape[1]
        full_nplet = range(N)
        for order in range(3,11):
            with self.subTest(order=order):
                nplets = torch.tensor(list(combinations(full_nplet, order)))
                batch_size = nplets.shape[0]
                nplets_hot_encoded = torch.zeros((batch_size, N), dtype=torch.int)
                nplets_hot_encoded[torch.arange(0,batch_size, dtype=int).view(-1,1), nplets] = 1
                res = nplets_measures_hot_encoded(self.covmat, nplets_hot_encoded, covmat_precomputed=True, T=self.X.shape[0])
                self._compare_with_ground_truth(res, nplets, rtol=1e-8, atol=1e-4)
    
    def test_multiple_times_same_dataset_timeseries_hot_encoded(self):
        N = self.X.shape[1]
        full_nplet = range(self.X.shape[1])
        for order in range(3,11):
            with self.subTest(order=order):
                nplets = torch.tensor(list(combinations(full_nplet, order)))
                batch_size = nplets.shape[0]
                nplets_hot_encoded = torch.zeros((batch_size, N), dtype=torch.int)
                nplets_hot_encoded[torch.arange(0,batch_size, dtype=int).view(-1,1), nplets] = 1
                res = nplets_measures_hot_encoded([self.X, self.X], nplets_hot_encoded)
                self._validate_same_results_for_repeated_datasets(res, nplets, rtol=1e-8, atol=1e-4)

    def test_multiple_times_same_dataset_precomputed_hot_encoded(self):
        N = self.X.shape[1]
        full_nplet = range(self.X.shape[1])
        for order in range(3,11):
            with self.subTest(order=order):
                nplets = torch.tensor(list(combinations(full_nplet, order)))
                batch_size = nplets.shape[0]
                nplets_hot_encoded = torch.zeros((batch_size, N), dtype=torch.int)
                nplets_hot_encoded[torch.arange(0,batch_size, dtype=int).view(-1,1), nplets] = 1
                res = nplets_measures_hot_encoded([self.covmat, self.covmat], nplets_hot_encoded, covmat_precomputed=True, T=self.X.shape[0])
                self._validate_same_results_for_repeated_datasets(res, nplets, rtol=1e-8, atol=1e-4)

    def test_batch_size_does_not_change_result_hot_encoded(self):
        full_nplet = range(self.X.shape[1])
        
        nplets = torch.tensor([list(c) for i, c in enumerate(combinations(full_nplet, 3)) if i < 100000])
        nplets_hot_encoded = torch.zeros((nplets.shape[0], self.X.shape[1]), dtype=torch.int)
        nplets_hot_encoded[torch.arange(0,nplets.shape[0], dtype=int).view(-1,1), nplets] = 1
        
        # test for different batch sizes
        res = nplets_measures_hot_encoded(self.X, nplets_hot_encoded)
        res2 = nplets_measures_hot_encoded(self.X, nplets_hot_encoded, batch_size=10)
        res3 = nplets_measures_hot_encoded(self.X, nplets_hot_encoded, batch_size=100)
        res4 = nplets_measures_hot_encoded(self.X, nplets_hot_encoded, batch_size=1000)
        res5 = nplets_measures_hot_encoded(self.X, nplets_hot_encoded, batch_size=10000)
        res6 = nplets_measures_hot_encoded(self.X, nplets_hot_encoded, batch_size=100000)
        res7 = nplets_measures_hot_encoded(self.X, nplets_hot_encoded, batch_size=1000000) # this should do a single batch
        
        # check that the results are the same
        self.assertTrue(torch.allclose(res, res2, rtol=1e-16, atol=1e-12))
        self.assertTrue(torch.allclose(res, res3, rtol=1e-16, atol=1e-12))
        self.assertTrue(torch.allclose(res, res4, rtol=1e-16, atol=1e-12))
        self.assertTrue(torch.allclose(res, res5, rtol=1e-16, atol=1e-12))
        self.assertTrue(torch.allclose(res, res6, rtol=1e-16, atol=1e-12))
        self.assertTrue(torch.allclose(res, res7, rtol=1e-16, atol=1e-12))

        
if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

# Read the contents of README.md
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='thoi',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    description='Torch - Higher Order Interactions',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Laouen Mayal Louan Belloli, Ruben Herzog',
    author_email='laouen.belloli@gmail.com',
    url='https://github.com/Laouen/THOI',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'tqdm',
        'scipy',
        'pandas',
        'networkx'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Programming Language :: Python :: 3'
    ],
    python_requires='>=3.6',
)


import torch


def _get_valid_candidates(initial_solution: torch.Tensor, N: int, device: torch.device) -> torch.Tensor:
    """
    Efficiently generates valid candidates by excluding elements present in the initial solution.

    Parameters:
    - initial_solution (torch.Tensor): Tensor of shape (batch_size, o) containing selected indices. Note: o < N and all elements must be unique.
    - N (int): Total number of possible elements.
    - device (torch.device): Device where tensors are located.

    Returns:
    - valid_candidates (torch.Tensor): Tensor of shape (batch_size, N - o) containing valid candidate indices.
    """

    batch_size, o = initial_solution.shape

    # Create a tensor of all elements, expanded across the batch
    # |batch_size| x |N|
    all_elements = torch.arange(N, device=device).unsqueeze(0).expand(batch_size, N)

    # Expand initial_solution for comparison
    # |batch_size| x |1| x |o|
    initial_solution_expanded = initial_solution.unsqueeze(1)

    # Expand all_elements for comparison
    # |batch_size| x |N| x |1|
    all_elements_expanded = all_elements.unsqueeze(2)

    # Compare all_elements with initial_solution
    # This creates a mask where True indicates the element is in initial_solution
    # |batch_size| x |N| x |o|
    is_in = (all_elements_expanded == initial_solution_expanded)

    # Determine if any of the o elements match for each (batch, N)
    # |batch_size| x |N|
    is_in_any = is_in.any(dim=2)

    # Create a mask for valid candidates (not in initial_solution)
    # |batch_size| x |N|
    valid_mask = ~is_in_any

    # Select valid candidates using the mask
    # |batch_size| x |N - o|
    valid_candidates = torch.masked_select(all_elements, valid_mask).view(batch_size, N - o)

    return valid_candidates


from typing import List, Union, Optional
import numpy as np
import scipy as sp
import torch

from thoi.typing import TensorLikeArray


def _get_string_metric(batched_res: np.ndarray, metric:str):
    '''
    Get the metric from the batched results returning the average over the D axis.

    params:
    - batched_res (np.ndarray): The batched results with shape (batch_size, D, 4) where 4 is the number of metrics (tc, dtc, o, s)
    - metric (str): The metric to get. One of tc, dtc, o or s
    '''

    METRICS = ['tc', 'dtc', 'o', 's']
    metric_idx = METRICS.index(metric)

    # |batch_size|
    return batched_res[:,:,metric_idx].mean(axis=1)

def gaussian_copula(X: np.ndarray):
    """
    Transform the data into a Gaussian copula and compute the covariance matrix.

    Parameters:
    - X: A 2D numpy array of shape (T, N) where T is the number of samples and N is the number of variables.

    Returns:
    - X_gaussian: The data transformed into the Gaussian copula (same shape as the parameter input).
    - X_gaussian_covmat: The covariance matrix of the Gaussian copula transformed data.
    """

    assert X.ndim == 2, f'data must be 2D but got {X.ndim}D data input'

    T = X.shape[0]

    # Step 1 & 2: Rank the data and normalize the ranks
    sortid = np.argsort(X, axis=0) # sorting indices
    copdata = np.argsort(sortid, axis=0) # sorting sorting indices
    copdata = (copdata+1)/(T+1) # normalized indices in the [0,1] range 

    # Step 3: Apply the inverse CDF of the standard normal distribution
    X_gaussian = sp.special.ndtri(copdata) #uniform data to gaussian

    # Handle infinite values by setting them to 0 (optional and depends on use case)
    X_gaussian[np.isinf(X_gaussian)] = 0

    # Step 4: Compute the covariance matrix
    X_gaussian_covmat = np.cov(X_gaussian.T)

    return X_gaussian, X_gaussian_covmat

def gaussian_copula_covmat(X: np.ndarray):
    return gaussian_copula(X)[1]

def _to_numpy(X):
    if isinstance(X, torch.Tensor):
        # If the tensor is on a GPU/TPU, move it to CPU first, then convert to NumPy
        return X.detach().cpu().numpy()
    elif isinstance(X, np.ndarray):
        return X
    return np.array(X)

def _get_device(use_cpu:bool=False):
    """Set the use of GPU if available"""
    using_GPU = torch.cuda.is_available() and not use_cpu
    device = torch.device('cuda' if using_GPU else 'cpu')
    return device

def _normalize_input_data(X: TensorLikeArray,
                         covmat_precomputed: bool=False,
                         T: Optional[Union[int, List[int]]]=None,
                         device: torch.device=torch.device('cpu')):
    '''
    brief: Normalize the input data to be a list of covariance matrices with shape (D, N, N) where D is the lenght of the list and N is the number of variables in the system.

    Parameters:
    - X: A list or a single 2D numpy arrays or tensors of shape: 1. (T, N) where T is the number of samples if X are multivariate series. 2. a list of 2D covariance matrices with shape (N, N).
    - covmat_precomputed: A boolean flag to indicate if the input data is a list of covariance matrices or multivariate series.
    - T (optional): A list of integers indicating the number of samples for each multivariate series.
    - device: The device to use for the computation. Default is 'cpu'.
    '''

    # Handle different options for X parameter. Accept multivariate data or covariance matrix
    if covmat_precomputed:
        covmats = torch.as_tensor(X)
        covmats = covmats.unsqueeze(0) if len(covmats.shape) == 2 else covmats
        assert covmats.shape[-2] == covmats.shape[-1], 'Covariance matrix should be square'
        assert len(covmats.shape) == 3, 'Covariance matrix should have dimensions (N, N) or (D, N, N)'
    else:
        
        try:
            X = _to_numpy(X)
            assert len(X.shape) in [2, 3], 'Covariance matrix should have dimensions (T, N) or (D, T, N)'
            X = [X] if len(X.shape) == 2 else [X[i] for i in range(X.shape[0])]
        except:
            X = [_to_numpy(x) for x in X]
            assert all([len(x.shape) == 2 for x in X]), 'All multivariate series should have dimensions (T, N) where T my vary and N be constant across all series'
            assert all([x.shape[1] == X[0].shape[1] for x in X]), 'All multivariate series should have dimensions (T, N) where T my vary and N be constant across all series'

        covmats = torch.stack([torch.from_numpy(gaussian_copula_covmat(x)) for x in X])
        T = [x.shape[0] for x in X]

    D, N = covmats.shape[:2]
    
    # Handle different options for T parameter
    if isinstance(T, int):
        T = [T] * D
    
    # Send covmat to device  
    covmats = covmats.to(device).contiguous()

    return covmats, D, N, T
